/*
 * Copyright (c) 2015-2016 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#ifndef _SYNTHPOD_PRIVATE_H
#define _SYNTHPOD_PRIVATE_H

#define SYNTHPOD_WORLD				SYNTHPOD_PREFIX"world"

#if defined(HAS_BUILTIN_ASSUME_ALIGNED)
#	define ASSUME_ALIGNED(PTR) __builtin_assume_aligned((PTR), 8)
#else
#	define ASSUME_ALIGNED(PTR) (PTR)
#endif

#include <lilv/lilv.h>

#include <lv2/lv2plug.in/ns/lv2core/lv2.h>
#include <lv2/lv2plug.in/ns/ext/urid/urid.h>
#include <lv2/lv2plug.in/ns/ext/midi/midi.h>
#include <lv2/lv2plug.in/ns/ext/log/log.h>
#include <lv2/lv2plug.in/ns/ext/worker/worker.h>
#include <lv2/lv2plug.in/ns/ext/buf-size/buf-size.h>
#include <lv2/lv2plug.in/ns/ext/resize-port/resize-port.h>
#include <lv2/lv2plug.in/ns/ext/presets/presets.h>
#include <lv2/lv2plug.in/ns/ext/patch/patch.h>
#include <lv2/lv2plug.in/ns/ext/port-props/port-props.h>
#include <lv2/lv2plug.in/ns/ext/port-groups/port-groups.h>
#include <lv2/lv2plug.in/ns/ext/state/state.h>
#include <lv2/lv2plug.in/ns/ext/time/time.h>
#include <lv2/lv2plug.in/ns/ext/parameters/parameters.h>
#include <lv2/lv2plug.in/ns/ext/instance-access/instance-access.h>
#include <lv2/lv2plug.in/ns/ext/data-access/data-access.h>
#include <lv2/lv2plug.in/ns/extensions/ui/ui.h>
#include <lv2/lv2plug.in/ns/extensions/units/units.h>
#include <lv2_external_ui.h> // kxstudio kx-ui extension
#include <lv2_extensions.h> // ardour inline display

#include <osc.lv2/osc.h>
#include <canvas.lv2/canvas.h>

#include <synthpod_stat.h>

typedef enum _port_type_t port_type_t;
typedef enum _port_atom_type_t port_atom_type_t;
typedef enum _port_buffer_type_t port_buffer_type_t;
typedef enum _port_direction_t port_direction_t;
typedef enum _port_protocol_t port_protocol_t;

enum _port_type_t {
	PORT_TYPE_AUDIO,
	PORT_TYPE_CONTROL,
	PORT_TYPE_CV,
	PORT_TYPE_ATOM,

	PORT_TYPE_NUM
};

enum _port_atom_type_t {
	PORT_ATOM_TYPE_ALL		= 0,

	PORT_ATOM_TYPE_MIDI		= (1 << 0),
	PORT_ATOM_TYPE_OSC		= (1 << 1),
	PORT_ATOM_TYPE_TIME		= (1 << 2),
	PORT_ATOM_TYPE_PATCH	= (1 << 3),
	PORT_ATOM_TYPE_XPRESS	= (1 << 4)
};

enum _port_buffer_type_t {
	PORT_BUFFER_TYPE_NONE = 0,
	PORT_BUFFER_TYPE_SEQUENCE,

	PORT_BUFFER_TYPE_NUM
};

enum _port_direction_t {
	PORT_DIRECTION_INPUT = 0,
	PORT_DIRECTION_OUTPUT,

	PORT_DIRECTION_NUM
};

enum _port_protocol_t {
	PORT_PROTOCOL_FLOAT = 0,
	PORT_PROTOCOL_PEAK,
	PORT_PROTOCOL_ATOM,
	PORT_PROTOCOL_SEQUENCE,

	PORT_PROTOCOL_NUM
}; //TODO use this

typedef int32_t u_id_t;

static inline void
sp_regs_init(LilvNode **nodes, LilvWorld *world)
{
	for(unsigned i = 1; i < STAT_URID_MAX; i++)
	{
		switch(i)
		{
			case SYNTHPOD__state:
				// fall-through
			case SYNTHPOD__null:
			{
				nodes[i] = lilv_new_string(world, stat_uris[i]);
			} break;

			default:
			{
				nodes[i] = lilv_new_uri(world, stat_uris[i]);
			} break;
		}
	}
}

static inline void
sp_regs_deinit(LilvNode **nodes)
{
	for(unsigned i = 1; i < STAT_URID_MAX; i++)
	{
		lilv_node_free(nodes[i]);
	}
}

// non-rt
static LilvNodes *
_preset_reload(LilvWorld *world, LilvNode *nodes[], const LilvPlugin *plugin,
	LilvNodes *presets, const char *bndl)
{
	if(presets)
	{
		// unload resources for this module's presets
		LILV_FOREACH(nodes, itr, presets)
		{
			const LilvNode *preset = lilv_nodes_get(presets, itr);

			lilv_world_unload_resource(world, preset);
		}

		// free presets for this module
		lilv_nodes_free(presets);
	}

	LilvNode *bndl_node = lilv_new_file_uri(world, NULL, bndl);
	if(bndl_node)
	{
		// unload bundle (if it should already be loaded)
		lilv_world_unload_bundle(world, bndl_node);

		// load bundle
		lilv_world_load_bundle(world, bndl_node);
		lilv_node_free(bndl_node);
	}

	// discover presets for this module
	presets = lilv_plugin_get_related(plugin, nodes[PRESETS__Preset]);

	// load resources for this module's presets
	LILV_FOREACH(nodes, itr, presets)
	{
		const LilvNode *preset = lilv_nodes_get(presets, itr);

		lilv_world_load_resource(world, preset);
	}

	return presets;
}

static inline int
_signum(LV2_URID urid1, LV2_URID urid2)
{
	if(urid1 < urid2)
		return -1;
	else if(urid1 > urid2)
		return 1;

	return 0;
}

#endif // _SYNTHPOD_PRIVATE_H
